from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..http import HTTPClient


@dataclass(slots=True)
class Emoji:
    """Represents a custom emoji in a Fluxer guild."""

    id: int
    name: str
    animated: bool = False
    guild_id: int | None = None
    roles: list[int] = field(default_factory=list)
    managed: bool = False
    available: bool = True

    _http: HTTPClient | None = field(default=None, repr=False)

    @classmethod
    def from_data(
        cls,
        data: dict[str, Any],
        http: HTTPClient | None = None,
        *,
        guild_id: int | None = None,
    ) -> Emoji:
        return cls(
            id=int(data["id"]),
            name=data.get("name", ""),
            animated=data.get("animated", False),
            guild_id=guild_id or (int(data["guild_id"]) if data.get("guild_id") else None),
            roles=[int(role_id) for role_id in data.get("roles", [])],
            managed=data.get("managed", False),
            available=data.get("available", True),
            _http=http,
        )

    async def delete(self, *, reason: str | None = None) -> None:
        """Delete this emoji from its guild.

        Args:
            reason: Reason for deletion (shows in audit log)

        Raises:
            Forbidden: You don't have permission to delete emojis
            NotFound: Emoji doesn't exist
            HTTPException: Deleting the emoji failed
        """
        if not self._http:
            raise RuntimeError("Cannot delete emoji without HTTPClient")
        if not self.guild_id:
            raise RuntimeError("Cannot delete emoji without guild_id")

        await self._http.delete_guild_emoji(self.guild_id, self.id, reason=reason)

    def __str__(self) -> str:
        return f"<{'a' if self.animated else ''}:{self.name}:{self.id}>"



from __future__ import annotations

import enum

# =============================================================================
# Gateway Opcodes
# Fluxer mirrors Discord's opcode structure since it's wire-compatible.
# =============================================================================


class GatewayOpcode(enum.IntEnum):
    """Gateway opcodes define the type of payload being sent/received."""

    DISPATCH = 0  # Server -> Client: An event was dispatched
    HEARTBEAT = 1  # Bidirectional: Maintain connection / request heartbeat
    IDENTIFY = 2  # Client -> Server: Start a new session
    PRESENCE_UPDATE = 3  # Client -> Server: Update client presence/status
    VOICE_STATE_UPDATE = 4  # Client -> Server: Join/move/leave voice channels
    RESUME = 6  # Client -> Server: Resume a dropped connection
    RECONNECT = 7  # Server -> Client: Client should reconnect
    REQUEST_GUILD_MEMBERS = 8  # Client -> Server: Request guild member list
    INVALID_SESSION = 9  # Server -> Client: Session is invalid
    HELLO = 10  # Server -> Client: Sent on connect, contains heartbeat_interval
    HEARTBEAT_ACK = 11  # Server -> Client: Acknowledgement of heartbeat


# =============================================================================
# Gateway Intents
# Bit flags that tell the gateway which events you want to receive.
# =============================================================================


class Intents(enum.IntFlag):
    """Gateway intents control which events your bot receives.

    Usage:
        intents = Intents.GUILDS | Intents.GUILD_MESSAGES
        intents = Intents.default()
        intents = Intents.all()
    """

    GUILDS = 1 << 0
    GUILD_MEMBERS = 1 << 1
    GUILD_MODERATION = 1 << 2
    GUILD_EMOJIS_AND_STICKERS = 1 << 3
    GUILD_INTEGRATIONS = 1 << 4
    GUILD_WEBHOOKS = 1 << 5
    GUILD_INVITES = 1 << 6
    GUILD_VOICE_STATES = 1 << 7
    GUILD_PRESENCES = 1 << 8
    GUILD_MESSAGES = 1 << 9
    GUILD_MESSAGE_REACTIONS = 1 << 10
    GUILD_MESSAGE_TYPING = 1 << 11
    DIRECT_MESSAGES = 1 << 12
    DIRECT_MESSAGE_REACTIONS = 1 << 13
    DIRECT_MESSAGE_TYPING = 1 << 14
    MESSAGE_CONTENT = 1 << 15

    @classmethod
    def default(cls) -> Intents:
        """Returns a sensible default set of intents (excludes privileged ones)."""
        value = cls(0)
        for intent in cls:
            if intent not in (
                cls.GUILD_MEMBERS,
                cls.GUILD_PRESENCES,
                cls.MESSAGE_CONTENT,
            ):
                value |= intent
        return value

    @classmethod
    def all(cls) -> Intents:
        """Returns all intents enabled."""
        value = cls(0)
        for intent in cls:
            value |= intent
        return value

    @classmethod
    def none(cls) -> Intents:
        """Returns no intents."""
        return cls(0)


# =============================================================================
# Gateway Close Codes
# =============================================================================


class GatewayCloseCode(enum.IntEnum):
    """WebSocket close codes the Fluxer gateway may send."""

    UNKNOWN_ERROR = 4000
    UNKNOWN_OPCODE = 4001
    DECODE_ERROR = 4002
    NOT_AUTHENTICATED = 4003
    AUTHENTICATION_FAILED = 4004
    ALREADY_AUTHENTICATED = 4005
    INVALID_SEQ = 4007
    RATE_LIMITED = 4008
    SESSION_TIMED_OUT = 4009
    INVALID_SHARD = 4010
    SHARDING_REQUIRED = 4011
    INVALID_API_VERSION = 4012
    INVALID_INTENTS = 4013
    DISALLOWED_INTENTS = 4014

    @property
    def is_reconnectable(self) -> bool:
        """Whether the bot should attempt to reconnect after this close code."""
        non_reconnectable = {
            self.AUTHENTICATION_FAILED,
            self.INVALID_SHARD,
            self.SHARDING_REQUIRED,
            self.INVALID_API_VERSION,
            self.INVALID_INTENTS,
            self.DISALLOWED_INTENTS,
        }
        return self not in non_reconnectable


# =============================================================================
# Channel Types
# =============================================================================


class ChannelType(enum.IntEnum):
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4
    GUILD_ANNOUNCEMENT = 5



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any

from ..utils import snowflake_to_datetime

if TYPE_CHECKING:
    from ..http import HTTPClient


@dataclass(slots=True)
class Guild:
    """Represents a Fluxer guild (server/community)."""

    id: int
    name: str | None = None
    icon: str | None = None
    owner_id: int | None = None
    member_count: int | None = None
    unavailable: bool = False

    _http: HTTPClient | None = field(default=None, repr=False)

    @classmethod
    def from_data(cls, data: dict[str, Any], http: HTTPClient | None = None) -> Guild:
        return cls(
            id=int(data["id"]),
            name=data.get("name"),
            icon=data.get("icon"),
            owner_id=int(data["owner_id"]) if data.get("owner_id") else None,
            member_count=data.get("member_count"),
            unavailable=data.get("unavailable", False),
            _http=http,
        )

    @property
    def created_at(self) -> datetime:
        return snowflake_to_datetime(self.id)

    @property
    def icon_url(self) -> str | None:
        if self.icon:
            ext = "gif" if self.icon.startswith("a_") else "png"
            return f"https://fluxerusercontent.com/icons/{self.id}/{self.icon}.{ext}"
        return None

    async def fetch_emojis(self) -> list[Any]:
        """Fetch all emojis in this guild.

        Returns:
            List of Emoji objects
        """
        if not self._http:
            raise RuntimeError("Cannot fetch emojis without HTTPClient")

        from .emoji import Emoji

        data = await self._http.get_guild_emojis(self.id)
        # Pass guild_id when creating emojis since API doesn't always return it
        return [
            Emoji.from_data(emoji_data, self._http, guild_id=self.id)
            for emoji_data in data
        ]

    def __str__(self) -> str:
        return self.name or f"Guild({self.id})"



from __future__ import annotations

import asyncio
import logging
from typing import Any

import aiohttp

from .errors import http_exception_from_status

log = logging.getLogger(__name__)

BASE_URL = "https://api.fluxer.app/v1"


def _get_user_agent() -> str:
    """Get the user agent string with the current version."""
    from . import __version__

    return f"fluxer.py/{__version__} (https://github.com/akarealemil/fluxer.py)"


class Route:
    """Represents an API route. Used for rate limit bucketing.

    Usage:
        route = Route("GET", "/channels/{channel_id}/messages", channel_id="123")
        # route.url = "https://api.fluxer.app/v1/channels/123/messages"
        # route.bucket = "GET /channels/{channel_id}/messages"
    """

    def __init__(self, method: str, path: str, **params: Any) -> None:
        self.method = method
        self.path = path
        # Convert all parameters to strings for URL formatting (handles int IDs)
        self.params = {k: str(v) for k, v in params.items()}
        self.url = BASE_URL + path.format(**self.params)

        # Rate limit bucket key: method + path template + major params
        # Major params (channel_id, guild_id) get their own buckets
        self.bucket = f"{method} {path}"
        for key in ("channel_id", "guild_id", "webhook_id"):
            if key in self.params:
                self.bucket += f":{self.params[key]}"


class RateLimiter:
    """Per-route rate limit handler using Fluxer's response headers.

    Fluxer returns rate limit info via HTTP headers (same pattern as Discord):
        X-RateLimit-Limit: max requests in window
        X-RateLimit-Remaining: requests left
        X-RateLimit-Reset: Unix timestamp when the limit resets
        X-RateLimit-Reset-After: seconds until reset
        X-RateLimit-Bucket: opaque bucket identifier
    """

    def __init__(self) -> None:
        self._locks: dict[str, asyncio.Lock] = {}
        self._reset_times: dict[str, float] = {}
        self._global_lock = asyncio.Event()
        self._global_lock.set()  # Not locked initially

    def _get_lock(self, bucket: str) -> asyncio.Lock:
        if bucket not in self._locks:
            self._locks[bucket] = asyncio.Lock()
        return self._locks[bucket]

    async def acquire(self, bucket: str) -> None:
        """Wait if this bucket or global rate limit is active."""
        # Wait for global rate limit to clear
        await self._global_lock.wait()

        lock = self._get_lock(bucket)
        await lock.acquire()

        # Check if we need to wait for this bucket
        reset_at = self._reset_times.get(bucket)
        if reset_at is not None:
            now = asyncio.get_event_loop().time()
            if now < reset_at:
                delay = reset_at - now
                log.debug("Rate limited on bucket %s, waiting %.2fs", bucket, delay)
                await asyncio.sleep(delay)

    def release(self, bucket: str, headers: dict[str, str]) -> None:
        """Update rate limit state from response headers and release the lock."""
        remaining = headers.get("X-RateLimit-Remaining")
        reset_after = headers.get("X-RateLimit-Reset-After")

        if remaining is not None and int(remaining) == 0 and reset_after is not None:
            delay = float(reset_after)
            self._reset_times[bucket] = asyncio.get_event_loop().time() + delay
            log.debug("Bucket %s exhausted, reset in %.2fs", bucket, delay)
        else:
            self._reset_times.pop(bucket, None)

        lock = self._get_lock(bucket)
        if lock.locked():
            lock.release()

    def set_global(self, retry_after: float) -> None:
        """Activate a global rate limit."""
        self._global_lock.clear()
        log.warning("Global rate limit hit, pausing for %.2fs", retry_after)

        async def _unlock() -> None:
            await asyncio.sleep(retry_after)
            self._global_lock.set()

        asyncio.ensure_future(_unlock())


class HTTPClient:
    """Async HTTP client for the Fluxer REST API.

    Usage:
        async with HTTPClient(token) as http:
            data = await http.request(Route("GET", "/users/@me"))
    """

    def __init__(self, token: str, *, is_bot: bool = True) -> None:
        self.token = token
        self.is_bot = is_bot
        self._session: aiohttp.ClientSession | None = None
        self._rate_limiter = RateLimiter()

    async def _ensure_session(self) -> aiohttp.ClientSession:
        if self._session is None or self._session.closed:
            # Use "Bot" prefix for bot tokens, plain token for user tokens
            auth_header = f"Bot {self.token}" if self.is_bot else self.token
            self._session = aiohttp.ClientSession(
                headers={
                    "Authorization": auth_header,
                    "User-Agent": _get_user_agent(),
                }
            )
        return self._session

    async def close(self) -> None:
        """Close the underlying HTTP session."""
        if self._session and not self._session.closed:
            await self._session.close()

    async def __aenter__(self) -> HTTPClient:
        await self._ensure_session()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()

    async def request(
        self,
        route: Route,
        *,
        json: Any = None,
        data: aiohttp.FormData | None = None,
        params: dict[str, Any] | None = None,
        reason: str | None = None,
    ) -> Any:
        """Make an authenticated request to the Fluxer API.

        Handles rate limiting, retries on 429/5xx, and error mapping.

        Returns:
            Parsed JSON response, or None for 204 No Content.
        """
        session = await self._ensure_session()

        headers: dict[str, str] = {}
        if reason:
            headers["X-Audit-Log-Reason"] = reason
        if json is not None:
            headers["Content-Type"] = "application/json"

        for attempt in range(5):  # Max retries
            await self._rate_limiter.acquire(route.bucket)

            try:
                async with session.request(
                    route.method,
                    route.url,
                    json=json,
                    data=data,
                    params=params,
                    headers=headers,
                ) as resp:
                    resp_headers = {k: v for k, v in resp.headers.items()}
                    self._rate_limiter.release(route.bucket, resp_headers)

                    # Success
                    if 200 <= resp.status < 300:
                        if resp.status == 204:
                            return None
                        return await resp.json()

                    # Rate limited
                    if resp.status == 429:
                        body = await resp.json()
                        retry_after = body.get("retry_after", 1.0)
                        is_global = body.get("global", False)

                        if is_global:
                            self._rate_limiter.set_global(retry_after)
                        else:
                            log.warning(
                                "Rate limited on %s, retry in %.2fs (attempt %d)",
                                route.url,
                                retry_after,
                                attempt + 1,
                            )
                            await asyncio.sleep(retry_after)
                        continue

                    # Server error — retry
                    if resp.status >= 500:
                        log.warning(
                            "Server error %d on %s, retrying (attempt %d)",
                            resp.status,
                            route.url,
                            attempt + 1,
                        )
                        await asyncio.sleep(1 + attempt)
                        continue

                    # Client error — raise
                    body = await resp.json()
                    raise http_exception_from_status(
                        status=resp.status,
                        code=body.get("code", "UNKNOWN"),
                        message=body.get("message", "Unknown error"),
                        errors=body.get("errors"),
                        retry_after=body.get("retry_after", 0.0),
                    )

            except (aiohttp.ClientError, asyncio.TimeoutError) as exc:
                if attempt < 4:
                    log.warning(
                        "Connection error: %s, retrying (attempt %d)", exc, attempt + 1
                    )
                    await asyncio.sleep(1 + attempt)
                    continue
                raise

        raise RuntimeError(f"Failed after 5 attempts: {route.method} {route.url}")

    # =========================================================================
    # Convenience methods for common endpoints
    # =========================================================================

    # -- Gateway --
    async def get_gateway(self) -> dict[str, Any]:
        """GET /gateway/bot — get the WebSocket URL.

        Note: Fluxer's /gateway endpoint returns 404 for bots.
        This method uses /gateway/bot instead (bot tokens only).
        """
        return await self.get_gateway_bot()

    async def get_gateway_bot(self) -> dict[str, Any]:
        """GET /gateway/bot — get gateway URL + sharding info."""
        return await self.request(Route("GET", "/gateway/bot"))

    # -- Users --
    async def get_current_user(self) -> dict[str, Any]:
        """GET /users/@me"""
        return await self.request(Route("GET", "/users/@me"))

    async def get_user(self, user_id: int | str) -> dict[str, Any]:
        """GET /users/{user_id}"""
        return await self.request(Route("GET", "/users/{user_id}", user_id=user_id))

    async def get_user_profile(
        self, user_id: int | str, *, guild_id: int | str | None = None
    ) -> dict[str, Any]:
        """GET /users/{user_id}/profile — Get a user's full profile.

        This returns additional profile information like bio, pronouns, banner, etc.
        that is not included in the basic user object.

        Args:
            user_id: The user ID to fetch
            guild_id: Optional guild ID for guild-specific profile data

        Returns:
            Profile object containing:
            - user: Basic user object
            - user_profile: Profile data (bio, pronouns, banner, etc.)
            - premium_type, premium_since, premium_lifetime_sequence
        """
        route = Route("GET", "/users/{user_id}/profile", user_id=user_id)
        params = {"guild_id": str(guild_id)} if guild_id else None
        return await self.request(route, params=params)

    async def get_current_user_guilds(self) -> list[dict[str, Any]]:
        """GET /users/@me/guilds - get guilds the current user is in"""
        return await self.request(Route("GET", "/users/@me/guilds"))

    # -- Channels --
    async def get_channel(self, channel_id: int | str) -> dict[str, Any]:
        """GET /channels/{channel_id}"""
        return await self.request(
            Route("GET", "/channels/{channel_id}", channel_id=channel_id)
        )

    # -- Messages --
    async def send_message(
        self,
        channel_id: int | str,
        *,
        content: str | None = None,
        embeds: list[dict[str, Any]] | None = None,
        files: list[Any] | None = None,
    ) -> dict[str, Any]:
        """POST /channels/{channel_id}/messages"""
        route = Route("POST", "/channels/{channel_id}/messages", channel_id=channel_id)

        payload: dict[str, Any] = {}
        if content is not None:
            payload["content"] = content
        if embeds is not None:
            payload["embeds"] = embeds

        if files:
            # Use multipart form data for file uploads
            form = aiohttp.FormData()
            import json as json_mod

            form.add_field(
                "payload_json", json_mod.dumps(payload), content_type="application/json"
            )
            for i, file in enumerate(files):
                form.add_field(f"files[{i}]", file["data"], filename=file["filename"])
            return await self.request(route, data=form)

        return await self.request(route, json=payload)

    async def get_messages(
        self,
        channel_id: int | str,
        *,
        limit: int = 50,
        before: int | str | None = None,
        after: int | str | None = None,
    ) -> list[dict[str, Any]]:
        """GET /channels/{channel_id}/messages"""
        params: dict[str, Any] = {"limit": limit}
        if before:
            params["before"] = before
        if after:
            params["after"] = after

        route = Route("GET", "/channels/{channel_id}/messages", channel_id=channel_id)
        return await self.request(route, params=params)

    async def edit_message(
        self,
        channel_id: int | str,
        message_id: int | str,
        *,
        content: str | None = None,
        embeds: list[dict[str, Any]] | None = None,
    ) -> dict[str, Any]:
        """PATCH /channels/{channel_id}/messages/{message_id}"""
        route = Route(
            "PATCH",
            "/channels/{channel_id}/messages/{message_id}",
            channel_id=channel_id,
            message_id=message_id,
        )
        payload: dict[str, Any] = {}
        if content is not None:
            payload["content"] = content
        if embeds is not None:
            payload["embeds"] = embeds
        return await self.request(route, json=payload)

    async def delete_message(self, channel_id: int | str, message_id: int | str) -> None:
        """DELETE /channels/{channel_id}/messages/{message_id}"""
        route = Route(
            "DELETE",
            "/channels/{channel_id}/messages/{message_id}",
            channel_id=channel_id,
            message_id=message_id,
        )
        await self.request(route)

    # -- Guilds --
    async def get_guild(self, guild_id: int | str) -> dict[str, Any]:
        """GET /guilds/{guild_id}"""
        return await self.request(Route("GET", "/guilds/{guild_id}", guild_id=guild_id))

    async def get_guild_channels(self, guild_id: int | str) -> list[dict[str, Any]]:
        """GET /guilds/{guild_id}/channels"""
        return await self.request(
            Route("GET", "/guilds/{guild_id}/channels", guild_id=guild_id)
        )

    async def get_guild_member(
        self, guild_id: int | str, user_id: int | str
    ) -> dict[str, Any]:
        """GET /guilds/{guild_id}/members/{user_id} — Get a specific guild member."""
        return await self.request(
            Route(
                "GET",
                "/guilds/{guild_id}/members/{user_id}",
                guild_id=guild_id,
                user_id=user_id,
            )
        )

    async def get_guild_members(
        self, guild_id: int | str, *, limit: int = 100, after: int | str | None = None
    ) -> list[dict[str, Any]]:
        """GET /guilds/{guild_id}/members — List guild members."""
        params: dict[str, Any] = {"limit": limit}
        if after:
            params["after"] = after

        return await self.request(
            Route("GET", "/guilds/{guild_id}/members", guild_id=guild_id), params=params
        )

    async def create_guild(
        self,
        *,
        name: str,
        icon: bytes | None = None,
    ) -> dict[str, Any]:
        """POST /guilds — Create a new guild.

        Args:
            name: Guild name (2-100 characters)
            icon: Icon image data (PNG/JPG/GIF)

        Returns:
            Guild object
        """
        import base64

        payload: dict[str, Any] = {"name": name}

        if icon:
            # Convert bytes to base64 data URI
            image_data = base64.b64encode(icon).decode("ascii")
            # Detect image format from header
            if icon.startswith(b"\x89PNG"):
                mime_type = "image/png"
            elif icon.startswith(b"\xff\xd8\xff"):
                mime_type = "image/jpeg"
            elif icon.startswith(b"GIF89a") or icon.startswith(b"GIF87a"):
                mime_type = "image/gif"
            else:
                mime_type = "image/png"  # Default

            payload["icon"] = f"data:{mime_type};base64,{image_data}"

        return await self.request(Route("POST", "/guilds"), json=payload)

    async def delete_guild(self, guild_id: int | str) -> None:
        """DELETE /guilds/{guild_id}"""
        await self.request(Route("DELETE", "/guilds/{guild_id}", guild_id=guild_id))

    async def modify_guild(
        self,
        guild_id: int | str,
        *,
        name: str | None = None,
        icon: bytes | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """PATCH /guilds/{guild_id} — Modify guild settings."""
        import base64

        payload: dict[str, Any] = {}

        if name is not None:
            payload["name"] = name

        if icon is not None:
            image_data = base64.b64encode(icon).decode("ascii")
            if icon.startswith(b"\x89PNG"):
                mime_type = "image/png"
            elif icon.startswith(b"\xff\xd8\xff"):
                mime_type = "image/jpeg"
            else:
                mime_type = "image/png"

            payload["icon"] = f"data:{mime_type};base64,{image_data}"

        payload.update(kwargs)

        return await self.request(
            Route("PATCH", "/guilds/{guild_id}", guild_id=guild_id), json=payload
        )

    # -- Roles --
    async def get_guild_roles(self, guild_id: int | str) -> list[dict[str, Any]]:
        """GET /guilds/{guild_id}/roles"""
        return await self.request(
            Route("GET", "/guilds/{guild_id}/roles", guild_id=guild_id)
        )

    async def create_guild_role(
        self,
        guild_id: int | str,
        *,
        name: str | None = None,
        permissions: int | None = None,
        color: int = 0,
        hoist: bool = False,
        mentionable: bool = False,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """POST /guilds/{guild_id}/roles — Create a new role."""
        payload: dict[str, Any] = {
            "color": color,
            "hoist": hoist,
            "mentionable": mentionable,
        }

        if name is not None:
            payload["name"] = name
        if permissions is not None:
            payload["permissions"] = str(permissions)

        payload.update(kwargs)

        return await self.request(
            Route("POST", "/guilds/{guild_id}/roles", guild_id=guild_id), json=payload
        )

    async def modify_guild_role(
        self,
        guild_id: int | str,
        role_id: int | str,
        *,
        name: str | None = None,
        permissions: int | None = None,
        color: int | None = None,
        hoist: bool | None = None,
        mentionable: bool | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """PATCH /guilds/{guild_id}/roles/{role_id}"""
        payload: dict[str, Any] = {}

        if name is not None:
            payload["name"] = name
        if permissions is not None:
            payload["permissions"] = str(permissions)
        if color is not None:
            payload["color"] = color
        if hoist is not None:
            payload["hoist"] = hoist
        if mentionable is not None:
            payload["mentionable"] = mentionable

        payload.update(kwargs)

        return await self.request(
            Route(
                "PATCH",
                "/guilds/{guild_id}/roles/{role_id}",
                guild_id=guild_id,
                role_id=role_id,
            ),
            json=payload,
        )

    async def delete_guild_role(self, guild_id: int | str, role_id: int | str) -> None:
        """DELETE /guilds/{guild_id}/roles/{role_id}"""
        await self.request(
            Route(
                "DELETE",
                "/guilds/{guild_id}/roles/{role_id}",
                guild_id=guild_id,
                role_id=role_id,
            )
        )

    # -- Channels (create/modify) --
    async def create_guild_channel(
        self,
        guild_id: int | str,
        *,
        name: str,
        type: int = 0,
        topic: str | None = None,
        bitrate: int | None = None,
        user_limit: int | None = None,
        position: int | None = None,
        parent_id: int | str | None = None,
        nsfw: bool = False,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """POST /guilds/{guild_id}/channels — Create a channel.

        Args:
            guild_id: Guild to create channel in
            name: Channel name
            type: Channel type (0=text, 2=voice, 4=category)
            topic: Channel topic (text channels)
            bitrate: Bitrate (voice channels)
            user_limit: User limit (voice channels)
            position: Channel position
            parent_id: Parent category ID
            nsfw: Whether the channel is NSFW

        Returns:
            Channel object
        """
        payload: dict[str, Any] = {
            "name": name,
            "type": type,
            "nsfw": nsfw,
        }

        if topic is not None:
            payload["topic"] = topic
        if bitrate is not None:
            payload["bitrate"] = bitrate
        if user_limit is not None:
            payload["user_limit"] = user_limit
        if position is not None:
            payload["position"] = position
        if parent_id is not None:
            payload["parent_id"] = parent_id

        payload.update(kwargs)

        return await self.request(
            Route("POST", "/guilds/{guild_id}/channels", guild_id=guild_id),
            json=payload,
        )

    async def modify_channel(
        self,
        channel_id: int | str,
        *,
        name: str | None = None,
        type: int | None = None,
        topic: str | None = None,
        position: int | None = None,
        parent_id: int | str | None = None,
        nsfw: bool | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """PATCH /channels/{channel_id}"""
        payload: dict[str, Any] = {}

        if name is not None:
            payload["name"] = name
        if type is not None:
            payload["type"] = type
        if topic is not None:
            payload["topic"] = topic
        if position is not None:
            payload["position"] = position
        if parent_id is not None:
            payload["parent_id"] = parent_id
        if nsfw is not None:
            payload["nsfw"] = nsfw

        payload.update(kwargs)

        return await self.request(
            Route("PATCH", "/channels/{channel_id}", channel_id=channel_id),
            json=payload,
        )

    async def delete_channel(self, channel_id: int | str) -> None:
        """DELETE /channels/{channel_id}"""
        await self.request(
            Route("DELETE", "/channels/{channel_id}", channel_id=channel_id)
        )

    async def edit_channel_permissions(
        self,
        channel_id: int | str,
        overwrite_id: int | str,
        *,
        allow: int | str | None = None,
        deny: int | str | None = None,
        type: int = 0,
        **kwargs: Any,
    ) -> None:
        """PUT /channels/{channel_id}/permissions/{overwrite_id} — Edit channel permission overwrites.

        Args:
            channel_id: Channel ID
            overwrite_id: Role or user ID
            allow: Allowed permissions (bitwise)
            deny: Denied permissions (bitwise)
            type: 0 for role, 1 for member

        Returns:
            None (204 No Content)
        """
        payload: dict[str, Any] = {"type": type}

        if allow is not None:
            payload["allow"] = str(allow)
        if deny is not None:
            payload["deny"] = str(deny)

        payload.update(kwargs)

        await self.request(
            Route(
                "PUT",
                "/channels/{channel_id}/permissions/{overwrite_id}",
                channel_id=channel_id,
                overwrite_id=overwrite_id,
            ),
            json=payload,
        )

    # -- User Profile --
    async def modify_current_user(
        self,
        *,
        username: str | None = None,
        avatar: bytes | None = None,
        banner: bytes | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """PATCH /users/@me — Modify the current user's profile.

        Args:
            username: New username
            avatar: Avatar image data (PNG/JPG/GIF)
            banner: Banner image data (PNG/JPG/GIF)

        Returns:
            Updated user object
        """
        import base64

        payload: dict[str, Any] = {}

        if username is not None:
            payload["username"] = username

        if avatar is not None:
            image_data = base64.b64encode(avatar).decode("ascii")
            if avatar.startswith(b"\x89PNG"):
                mime_type = "image/png"
            elif avatar.startswith(b"\xff\xd8\xff"):
                mime_type = "image/jpeg"
            elif avatar.startswith(b"GIF89a") or avatar.startswith(b"GIF87a"):
                mime_type = "image/gif"
            else:
                mime_type = "image/png"

            payload["avatar"] = f"data:{mime_type};base64,{image_data}"

        if banner is not None:
            image_data = base64.b64encode(banner).decode("ascii")
            if banner.startswith(b"\x89PNG"):
                mime_type = "image/png"
            elif banner.startswith(b"\xff\xd8\xff"):
                mime_type = "image/jpeg"
            elif banner.startswith(b"GIF89a") or banner.startswith(b"GIF87a"):
                mime_type = "image/gif"
            else:
                mime_type = "image/png"

            payload["banner"] = f"data:{mime_type};base64,{image_data}"

        payload.update(kwargs)

        return await self.request(Route("PATCH", "/users/@me"), json=payload)

    # -- Emojis --
    async def get_guild_emojis(self, guild_id: int | str) -> list[dict[str, Any]]:
        """GET /guilds/{guild_id}/emojis — Get all emojis for a guild.

        Returns:
            List of emoji objects
        """
        return await self.request(
            Route("GET", "/guilds/{guild_id}/emojis", guild_id=guild_id)
        )

    async def get_guild_emoji(self, guild_id: int | str, emoji_id: int | str) -> dict[str, Any]:
        """GET /guilds/{guild_id}/emojis/{emoji_id} — Get a specific emoji.

        Returns:
            Emoji object
        """
        return await self.request(
            Route(
                "GET",
                "/guilds/{guild_id}/emojis/{emoji_id}",
                guild_id=guild_id,
                emoji_id=emoji_id,
            )
        )

    async def create_guild_emoji(
        self,
        guild_id: int | str,
        *,
        name: str,
        image: bytes,
        roles: list[int | str] | None = None,
        reason: str | None = None,
    ) -> dict[str, Any]:
        """POST /guilds/{guild_id}/emojis — Create a new emoji.

        Args:
            guild_id: Guild ID
            name: Emoji name
            image: Image data (PNG/JPG/GIF)
            roles: List of role IDs that can use this emoji (optional)
            reason: Reason for creation (audit log)

        Returns:
            Emoji object
        """
        import base64

        # Convert bytes to base64 data URI
        image_data = base64.b64encode(image).decode("ascii")

        # Detect image format from header
        if image.startswith(b"\x89PNG"):
            mime_type = "image/png"
        elif image.startswith(b"\xff\xd8\xff"):
            mime_type = "image/jpeg"
        elif image.startswith(b"GIF89a") or image.startswith(b"GIF87a"):
            mime_type = "image/gif"
        else:
            mime_type = "image/png"  # Default

        payload: dict[str, Any] = {
            "name": name,
            "image": f"data:{mime_type};base64,{image_data}",
        }

        if roles is not None:
            payload["roles"] = [str(role_id) for role_id in roles]

        return await self.request(
            Route("POST", "/guilds/{guild_id}/emojis", guild_id=guild_id),
            json=payload,
            reason=reason,
        )

    async def delete_guild_emoji(
        self,
        guild_id: int | str,
        emoji_id: int | str,
        *,
        reason: str | None = None,
    ) -> None:
        """DELETE /guilds/{guild_id}/emojis/{emoji_id} — Delete an emoji.

        Args:
            guild_id: Guild ID
            emoji_id: Emoji ID
            reason: Reason for deletion (audit log)
        """
        await self.request(
            Route(
                "DELETE",
                "/guilds/{guild_id}/emojis/{emoji_id}",
                guild_id=guild_id,
                emoji_id=emoji_id,
            ),
            reason=reason,
        )

    # ~~ Webhooks ~
    async def get_guild_webhooks(self, guild_id: int | str) -> list[dict[str, Any]]:
        """GET /guilds/{guild_id}/webhooks"""
        return await self.request(
            Route("GET", "/guilds/{guild_id}/webhooks", guild_id=guild_id)
        )

    async def get_channel_webhooks(self, channel_id: int | str) -> list[dict[str, Any]]:
        """GET /channels/{channel_id}/webhooks"""
        return await self.request(
            Route("GET", "/channels/{channel_id}/webhooks", channel_id=channel_id)
        )

    async def create_webhook(
        self,
        channel_id: int | str,
        *,
        name: str,
        avatar: str | None = None,
    ) -> dict[str, Any]:
        """POST /channels/{channel_id}/webhooks"""
        payload: dict[str, Any] = {"name": name}
        if avatar is not None:
            payload["avatar"] = avatar
        return await self.request(
            Route("POST", "/channels/{channel_id}/webhooks", channel_id=channel_id),
            json=payload,
        )

    async def get_webhook(self, webhook_id: int | str) -> dict[str, Any]:
        """GET /webhooks/{webhook_id}"""
        return await self.request(
            Route("GET", "/webhooks/{webhook_id}", webhook_id=webhook_id)
        )

    async def get_webhook_with_token(
        self, webhook_id: int | str, token: str
    ) -> dict[str, Any]:
        """GET /webhooks/{webhook_id}/{token}"""
        return await self.request(
            Route(
                "GET",
                "/webhooks/{webhook_id}/{token}",
                webhook_id=webhook_id,
                token=token,
            )
        )

    async def modify_webhook(
        self,
        webhook_id: int | str,
        *,
        name: str | None = None,
        avatar: str | None = None,
        channel_id: int | str | None = None,
    ) -> dict[str, Any]:
        """PATCH /webhooks/{webhook_id}"""
        payload: dict[str, Any] = {}
        if name is not None:
            payload["name"] = name
        if avatar is not None:
            payload["avatar"] = avatar
        if channel_id is not None:
            payload["channel_id"] = channel_id
        return await self.request(
            Route("PATCH", "/webhooks/{webhook_id}", webhook_id=webhook_id),
            json=payload,
        )

    async def modify_webhook_with_token(
        self,
        webhook_id: int | str,
        token: str,
        *,
        name: str | None = None,
        avatar: str | None = None,
        channel_id: int | str | None = None,
    ) -> dict[str, Any]:
        """PATCH /webhooks/{webhook_id}/{token}"""
        payload: dict[str, Any] = {}
        if name is not None:
            payload["name"] = name
        if avatar is not None:
            payload["avatar"] = avatar
        if channel_id is not None:
            payload["channel_id"] = channel_id
        return await self.request(
            Route(
                "PATCH",
                "/webhooks/{webhook_id}/{token}",
                webhook_id=webhook_id,
                token=token,
            ),
            json=payload,
        )

    async def delete_webhook(
        self, webhook_id: int | str, *, reason: str | None = None
    ) -> None:
        """DELETE /webhooks/{webhook_id}"""
        await self.request(
            Route("DELETE", "/webhooks/{webhook_id}", webhook_id=webhook_id),
            reason=reason,
        )

    async def delete_webhook_with_token(self, webhook_id: int | str, token: str) -> None:
        """DELETE /webhooks/{webhook_id}/{token}"""
        await self.request(
            Route(
                "DELETE",
                "/webhooks/{webhook_id}/{token}",
                webhook_id=webhook_id,
                token=token,
            ),
        )

    async def execute_webhook(
        self,
        webhook_id: int | str,
        token: str,
        *,
        content: str | None = None,
        embeds: list[dict[str, Any]] | None = None,
        username: str | None = None,
        avatar_url: str | None = None,
        wait: bool = False,
    ) -> dict[str, Any] | None:
        """POST /webhooks/{webhook_id}/{token}"""
        payload: dict[str, Any] = {}
        if content is not None:
            payload["content"] = content
        if embeds is not None:
            payload["embeds"] = embeds
        if username is not None:
            payload["username"] = username
        if avatar_url is not None:
            payload["avatar_url"] = avatar_url
        params = {"wait": "true"} if wait else None
        return await self.request(
            Route(
                "POST",
                "/webhooks/{webhook_id}/{token}",
                webhook_id=webhook_id,
                token=token,
            ),
            json=payload,
            params=params,
        )



"""Quick inspection test to see what fields Fluxer API returns.

Run with pytest:
    pytest tests/test_inspect_api.py -v -s

Or run directly:
    python tests/test_inspect_api.py
"""

import asyncio
import json

import pytest

from fluxer.http import HTTPClient


@pytest.mark.asyncio
async def test_inspect_user_fields(http_client: HTTPClient) -> None:
    """Print out all fields returned by the user endpoint (pytest version)."""
    data = await http_client.get_current_user()

    print("\n" + "=" * 60)
    print("User object fields from Fluxer API:")
    print("=" * 60)
    print(json.dumps(data, indent=2))
    print("=" * 60)

    # List all keys
    print(f"\nAvailable fields: {list(data.keys())}")
    print()

    # Assert we at least got the basic fields
    assert "id" in data
    assert "username" in data
    assert "discriminator" in data


async def inspect_user_fields() -> None:
    """Print out all fields returned by the user endpoint (standalone version)."""
    token = "bot_token"

    async with HTTPClient(token) as http:
        data = await http.get_current_user()

        print("\n" + "=" * 60)
        print("User object fields from Fluxer API:")
        print("=" * 60)
        print(json.dumps(data, indent=2))
        print("=" * 60)

        # List all keys
        print(f"\nAvailable fields: {list(data.keys())}")
        print()


if __name__ == "__main__":
    asyncio.run(inspect_user_fields())



from __future__ import annotations

from datetime import datetime, timezone

# Fluxer uses the same epoch as Discord: 2015-01-01T00:00:00Z
FLUXER_EPOCH = 1420070400000


def snowflake_to_datetime(snowflake: str | int) -> datetime:
    """Convert a Fluxer Snowflake ID to a datetime.

    Snowflakes encode a timestamp in the upper 42 bits.

    Args:
        snowflake: The Snowflake ID as a string or int.

    Returns:
        A timezone-aware UTC datetime.
    """
    timestamp_ms = (int(snowflake) >> 22) + FLUXER_EPOCH
    return datetime.fromtimestamp(timestamp_ms / 1000, tz=timezone.utc)


def datetime_to_snowflake(dt: datetime) -> int:
    """Convert a datetime to a Snowflake ID (useful for pagination).

    This creates a Snowflake with only the timestamp component set.
    Useful for before/after pagination parameters.

    Args:
        dt: A datetime object.

    Returns:
        A Snowflake integer.
    """
    timestamp_ms = int(dt.timestamp() * 1000)
    snowflake = (timestamp_ms - FLUXER_EPOCH) << 22
    return snowflake



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any

from ..utils import snowflake_to_datetime

if TYPE_CHECKING:
    from ..http import HTTPClient


@dataclass(slots=True)
class User:
    """Represents a Fluxer user.

    This contains fields that are available across most endpoints (guild members,
    message authors, etc.). Some fields like bio, banner, and banner_color are
    ONLY available when fetching your own user via GET /users/@me, not when
    fetching other users via GET /users/{user_id}.

    Fields that are ONLY in GET /users/@me:
    - bio: User's bio/about me text
    - banner: Banner image hash
    - banner_color: Banner color
    - email, phone: Contact info
    - premium fields, MFA status, etc.
    """

    # Core identity fields (available everywhere)
    id: int
    username: str
    discriminator: str | None = None
    global_name: str | None = None  # Display name/nickname
    avatar_hash: str | None = None
    avatar_color: str | None = None
    bot: bool = False
    flags: int = 0

    # Additional profile fields (may not be available for all users)
    bio: str | None = None  # User bio/about me
    banner_hash: str | None = None  # Banner image hash
    banner_color: int | None = None  # Banner color (integer)

    # Back-reference (set after construction)
    _http: HTTPClient | None = field(default=None, repr=False)

    @classmethod
    def from_data(cls, data: dict[str, Any], http: HTTPClient | None = None) -> User:
        return cls(
            id=int(data["id"]),
            username=data["username"],
            discriminator=data.get("discriminator"),
            global_name=data.get("global_name"),
            avatar_hash=data.get("avatar"),
            avatar_color=data.get("avatar_color"),
            bot=data.get("bot", False),
            flags=data.get("flags", 0),
            bio=data.get("bio"),
            banner_hash=data.get("banner"),
            banner_color=data.get("banner_color"),
            _http=http,
        )

    @property
    def created_at(self) -> datetime:
        """When this user account was created (derived from Snowflake)."""
        return snowflake_to_datetime(self.id)

    @property
    def display_name(self) -> str:
        """The best display name for this user.

        Returns global_name if set, otherwise falls back to username.
        This is the name you should show to users.
        """
        return self.global_name or self.username

    @property
    def mention(self) -> str:
        """Return a string that mentions this user in a message."""
        return f"<@{self.id}>"

    @property
    def avatar_url(self) -> str | None:
        """URL for the user's avatar, or None if they use the default."""
        if self.avatar_hash:
            ext = "gif" if self.avatar_hash.startswith("a_") else "png"
            return (
                f"https://fluxerusercontent.com/avatars/{self.id}/{self.avatar_hash}.{ext}"
            )
        return None

    @property
    def default_avatar_url(self) -> str:
        """URL for the user's default avatar."""
        index = int(self.id) % 6
        return f"https://fluxerstatic.com/avatars/{index}.png"

    @property
    def banner_url(self) -> str | None:
        """URL for the user's banner, or None if they don't have one."""
        if self.banner_hash:
            ext = "gif" if self.banner_hash.startswith("a_") else "png"
            return (
                f"https://fluxerusercontent.com/banners/{self.id}/{self.banner_hash}.{ext}"
            )
        return None

    def __eq__(self, other: object) -> bool:
        return isinstance(other, User) and self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    def __str__(self) -> str:
        """Return the user's display name.

        Uses global_name if available, otherwise username.
        """
        return self.global_name or self.username



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from .user import User

if TYPE_CHECKING:
    from ..http import HTTPClient


@dataclass(slots=True)
class UserProfile:
    """Represents a Fluxer user's full profile.

    This contains profile information that is only available via the
    GET /users/{id}/profile endpoint, not from basic user objects.
    """

    # The basic user object
    user: User

    # Profile information
    bio: str | None = None
    pronouns: str | None = None
    banner: str | None = None  # Banner image hash
    banner_color: int | None = None
    accent_color: int | None = None

    # Premium information
    premium_type: int | None = None
    premium_since: str | None = None  # ISO 8601 timestamp
    premium_lifetime_sequence: int | None = None

    # Back-reference (set after construction)
    _http: HTTPClient | None = field(default=None, repr=False)

    @classmethod
    def from_data(cls, data: dict[str, Any], http: HTTPClient | None = None) -> UserProfile:
        """Construct a UserProfile from raw API data.

        Args:
            data: Raw profile object from GET /users/{id}/profile
            http: HTTPClient for making further requests

        Returns:
            A new UserProfile instance
        """
        # Parse the nested user object
        user = User.from_data(data["user"], http)

        # Get the user_profile section
        profile_data = data.get("user_profile", {})

        return cls(
            user=user,
            bio=profile_data.get("bio"),
            pronouns=profile_data.get("pronouns"),
            banner=profile_data.get("banner"),
            banner_color=profile_data.get("banner_color"),
            accent_color=profile_data.get("accent_color"),
            premium_type=data.get("premium_type"),
            premium_since=data.get("premium_since"),
            premium_lifetime_sequence=data.get("premium_lifetime_sequence"),
            _http=http,
        )

    @property
    def banner_url(self) -> str | None:
        """URL for the user's banner, or None if they don't have one."""
        if self.banner:
            ext = "gif" if self.banner.startswith("a_") else "png"
            return (
                f"https://fluxerusercontent.com/banners/{self.user.id}/{self.banner}.{ext}"
            )
        return None

    @property
    def is_premium(self) -> bool:
        """Whether this user has premium."""
        return self.premium_type is not None and self.premium_type > 0

    def __str__(self) -> str:
        """Return the user's display name."""
        return self.user.display_name



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any

from ..utils import snowflake_to_datetime

if TYPE_CHECKING:
    from ..http import HTTPClient
    from .channel import Channel
    from .user import User


@dataclass(slots=True)
class Message:
    """Represents a message in a Fluxer channel."""

    id: int
    channel_id: int
    content: str
    author: User
    timestamp: str
    edited_timestamp: str | None = None
    guild_id: int | None = None
    embeds: list[dict[str, Any]] = field(default_factory=list)
    attachments: list[dict[str, Any]] = field(default_factory=list)
    mentions: list[User] = field(default_factory=list)
    pinned: bool = False

    _http: HTTPClient | None = field(default=None, repr=False)
    _channel: Channel | None = field(default=None, repr=False)

    @classmethod
    def from_data(cls, data: dict[str, Any], http: HTTPClient | None = None) -> Message:
        from .user import User

        author = User.from_data(data["author"], http)
        mentions = [User.from_data(u, http) for u in data.get("mentions", [])]

        return cls(
            id=int(data["id"]),
            channel_id=int(data["channel_id"]),
            content=data.get("content", ""),
            author=author,
            timestamp=data["timestamp"],
            edited_timestamp=data.get("edited_timestamp"),
            guild_id=int(data["guild_id"]) if data.get("guild_id") else None,
            embeds=data.get("embeds", []),
            attachments=data.get("attachments", []),
            mentions=mentions,
            pinned=data.get("pinned", False),
            _http=http,
        )

    @property
    def created_at(self) -> datetime:
        return snowflake_to_datetime(self.id)

    @property
    def channel(self) -> Channel | None:
        """The channel this message was sent in (if cached)."""
        return self._channel

    async def reply(self, content: str, **kwargs: Any) -> Message:
        """Reply to this message."""
        if self._http is None:
            raise RuntimeError("Message is not bound to an HTTP client")
        data = await self._http.send_message(self.channel_id, content=content, **kwargs)
        return Message.from_data(data, self._http)

    async def edit(self, content: str | None = None, **kwargs: Any) -> Message:
        """Edit this message."""
        if self._http is None:
            raise RuntimeError("Message is not bound to an HTTP client")
        data = await self._http.edit_message(
            self.channel_id, self.id, content=content, **kwargs
        )
        return Message.from_data(data, self._http)

    async def delete(self) -> None:
        """Delete this message."""
        if self._http is None:
            raise RuntimeError("Message is not bound to an HTTP client")
        await self._http.delete_message(self.channel_id, self.id)



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING, Any

from ..enums import ChannelType
from ..utils import snowflake_to_datetime

if TYPE_CHECKING:
    from ..http import HTTPClient
    from .embed import Embed
    from .message import Message


@dataclass(slots=True)
class Channel:
    """Represents a Fluxer channel (text, DM, voice, category, etc.)."""

    id: int
    type: int
    name: str | None = None
    guild_id: int | None = None
    position: int | None = None
    topic: str | None = None
    nsfw: bool = False
    parent_id: int | None = None

    _http: HTTPClient | None = field(default=None, repr=False)

    @classmethod
    def from_data(cls, data: dict[str, Any], http: HTTPClient | None = None) -> Channel:
        return cls(
            id=int(data["id"]),
            type=data["type"],
            name=data.get("name"),
            guild_id=int(data["guild_id"]) if data.get("guild_id") else None,
            position=data.get("position"),
            topic=data.get("topic"),
            nsfw=data.get("nsfw", False),
            parent_id=int(data["parent_id"]) if data.get("parent_id") else None,
            _http=http,
        )

    @property
    def mention(self) -> str:
        return f"<#{self.id}>"

    @property
    def created_at(self) -> datetime:
        return snowflake_to_datetime(self.id)

    @property
    def is_text_channel(self) -> bool:
        """Whether this is a guild text channel."""
        return self.type == ChannelType.GUILD_TEXT

    @property
    def is_voice_channel(self) -> bool:
        """Whether this is a voice channel."""
        return self.type == ChannelType.GUILD_VOICE

    @property
    def is_dm(self) -> bool:
        """Whether this is a DM channel."""
        return self.type == ChannelType.DM

    @property
    def is_category(self) -> bool:
        """Whether this is a category channel."""
        return self.type == ChannelType.GUILD_CATEGORY

    async def send(
        self,
        content: str | None = None,
        *,
        embed: Embed | None = None,
        embeds: list[Embed] | None = None,
    ) -> Message:
        """Send a message to this channel.

        Args:
            content: Text content of the message.
            embed: A single embed to include.
            embeds: Multiple embeds to include.

        Returns:
            The created Message object.
        """
        # Import here to avoid circular imports
        from .message import Message

        if self._http is None:
            raise RuntimeError("Channel is not bound to an HTTP client")

        embed_list: list[dict[str, Any]] | None = None
        if embed:
            embed_list = [embed.to_dict()]
        elif embeds:
            embed_list = [e.to_dict() for e in embeds]

        data = await self._http.send_message(
            self.id, content=content, embeds=embed_list
        )
        return Message.from_data(data, self._http)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Channel) and self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ..http import HTTPClient

from .user import User


@dataclass(slots=True)
class GuildMember:
    """Represents a member of a guild.

    This combines a User object with guild-specific information like
    nickname, roles, join date, etc.
    """

    # The underlying user
    user: User

    # Guild-specific fields
    nick: str | None = None  # Guild nickname (overrides global_name)
    avatar_hash: str | None = None  # Guild-specific avatar hash
    banner: str | None = None  # Guild-specific banner
    accent_color: int | None = None  # Guild-specific accent color
    roles: list[int] = field(default_factory=list)  # Role IDs (as ints)
    joined_at: str | None = None  # ISO 8601 timestamp

    # Invite/join tracking
    join_source_type: int | None = None
    source_invite_code: str | None = None
    inviter_id: int | None = None

    # Voice state
    mute: bool = False
    deaf: bool = False

    # Moderation
    communication_disabled_until: str | None = None  # Timeout until (ISO 8601)

    # Back-reference (set after construction)
    _http: HTTPClient | None = field(default=None, repr=False)

    @classmethod
    def from_data(
        cls, data: dict[str, Any], http: HTTPClient | None = None
    ) -> GuildMember:
        # Parse the nested user object
        user = User.from_data(data["user"], http)

        return cls(
            user=user,
            nick=data.get("nick"),
            avatar_hash=data.get("avatar"),
            banner=data.get("banner"),
            accent_color=data.get("accent_color"),
            roles=[int(role_id) for role_id in data.get("roles", [])],
            joined_at=data.get("joined_at"),
            join_source_type=data.get("join_source_type"),
            source_invite_code=data.get("source_invite_code"),
            inviter_id=int(data["inviter_id"]) if data.get("inviter_id") else None,
            mute=data.get("mute", False),
            deaf=data.get("deaf", False),
            communication_disabled_until=data.get("communication_disabled_until"),
            _http=http,
        )

    @property
    def display_name(self) -> str:
        """The best display name for this member.

        Priority: guild nickname > global name > username
        """
        return self.nick or self.user.global_name or self.user.username

    @property
    def mention(self) -> str:
        """Return a string that mentions this member."""
        return f"<@{self.user.id}>"

    @property
    def guild_avatar_url(self) -> str | None:
        """URL for the member's guild-specific avatar, if set."""
        if self.avatar_hash:
            ext = "gif" if self.avatar_hash.startswith("a_") else "png"
            # Note: Guild avatar URLs might have a different format
            # Adjust if Fluxer uses a different URL structure
            return f"https://fluxerusercontent.com/guilds/avatars/{self.user.id}/{self.avatar_hash}.{ext}"
        return None

    def __str__(self) -> str:
        """Return the member's display name."""
        return self.display_name



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any


@dataclass
class Embed:
    """Builder for rich embed objects.

    Usage:
        embed = Embed(title="Hello", description="World", color=0x5865F2)
        embed.add_field(name="Field 1", value="Value 1")
        embed.set_footer(text="Footer text")
        await channel.send(embed=embed)
    """

    title: str | None = None
    description: str | None = None
    url: str | None = None
    color: int | None = None
    timestamp: str | None = None
    footer: dict[str, Any] | None = None
    image: dict[str, Any] | None = None
    thumbnail: dict[str, Any] | None = None
    author: dict[str, Any] | None = None
    fields: list[dict[str, Any]] = field(default_factory=list)

    def set_footer(self, *, text: str, icon_url: str | None = None) -> Embed:
        self.footer = {"text": text}
        if icon_url:
            self.footer["icon_url"] = icon_url
        return self

    def set_image(self, *, url: str) -> Embed:
        self.image = {"url": url}
        return self

    def set_thumbnail(self, *, url: str) -> Embed:
        self.thumbnail = {"url": url}
        return self

    def set_author(
        self, *, name: str, url: str | None = None, icon_url: str | None = None
    ) -> Embed:
        self.author = {"name": name}
        if url:
            self.author["url"] = url
        if icon_url:
            self.author["icon_url"] = icon_url
        return self

    def add_field(self, *, name: str, value: str, inline: bool = False) -> Embed:
        self.fields.append({"name": name, "value": value, "inline": inline})
        return self

    def to_dict(self) -> dict[str, Any]:
        """Serialize to a dict suitable for the API."""
        d: dict[str, Any] = {}
        if self.title is not None:
            d["title"] = self.title
        if self.description is not None:
            d["description"] = self.description
        if self.url is not None:
            d["url"] = self.url
        if self.color is not None:
            d["color"] = self.color
        if self.timestamp is not None:
            d["timestamp"] = self.timestamp
        if self.footer is not None:
            d["footer"] = self.footer
        if self.image is not None:
            d["image"] = self.image
        if self.thumbnail is not None:
            d["thumbnail"] = self.thumbnail
        if self.author is not None:
            d["author"] = self.author
        if self.fields:
            d["fields"] = self.fields
        return d



from __future__ import annotations

import asyncio
import json
import logging
import sys
from typing import Any, Callable, Coroutine

import aiohttp

from .enums import GatewayCloseCode, GatewayOpcode, Intents
from .errors import GatewayNotConnected

log = logging.getLogger(__name__)


class GatewayPayload:
    """Represents a Gateway payload (sent or received).

    All gateway messages follow this structure:
        op: opcode (int)
        d:  event data (dict, int, or None)
        s:  sequence number (only for op 0 DISPATCH)
        t:  event name (only for op 0 DISPATCH)
    """

    __slots__ = ("op", "d", "s", "t")

    def __init__(
        self, op: int, d: Any = None, s: int | None = None, t: str | None = None
    ):
        self.op = op
        self.d = d
        self.s = s
        self.t = t

    @classmethod
    def from_json(cls, raw: str) -> GatewayPayload:
        data = json.loads(raw)
        return cls(op=data["op"], d=data.get("d"), s=data.get("s"), t=data.get("t"))

    def to_json(self) -> str:
        payload: dict[str, Any] = {"op": self.op, "d": self.d}
        if self.s is not None:
            payload["s"] = self.s
        if self.t is not None:
            payload["t"] = self.t
        return json.dumps(payload)

    def __repr__(self) -> str:
        op_name = (
            GatewayOpcode(self.op).name
            if self.op in GatewayOpcode.__members__.values()
            else str(self.op)
        )
        return f"<GatewayPayload op={op_name} t={self.t!r} s={self.s}>"


class Gateway:
    """Manages the WebSocket connection to the Fluxer Gateway.

    This class handles the full lifecycle: connect, heartbeat, identify,
    dispatch events, and reconnect on failure.
    """

    def __init__(
        self,
        *,
        http_client: Any,
        token: str,
        intents: Intents,
        dispatch: Callable[[str, Any], Coroutine[Any, Any, None]],
    ) -> None:
        self._http = http_client
        self._token = token
        self._intents = intents
        self._dispatch = dispatch

        # Connection state
        self._ws: aiohttp.ClientWebSocketResponse | None = None
        self._session: aiohttp.ClientSession | None = None
        self._heartbeat_interval: float = 41.25
        self._sequence: int | None = None
        self._session_id: str | None = None
        self._resume_gateway_url: str | None = None
        self._heartbeat_task: asyncio.Task[None] | None = None
        self._is_closed: bool = False
        self._last_heartbeat_ack: bool = True

        self._tasks = []

    @property
    def is_connected(self) -> bool:
        return self._ws is not None and not self._ws.closed

    async def connect(self) -> None:
        """Establish the gateway connection and enter the event loop."""
        gateway_data = await self._http.get_gateway_bot()
        gateway_url = gateway_data["url"]
        ws_url = f"{gateway_url}?v=1&encoding=json"

        log.info("Connecting to gateway: %s", ws_url)

        self._session = aiohttp.ClientSession()
        self._is_closed = False

        while not self._is_closed:
            try:
                await self._connect_and_run(ws_url)
            except (
                aiohttp.WSServerHandshakeError,
                aiohttp.ClientError,
                asyncio.TimeoutError,
            ) as e:
                log.warning("Gateway connection error: %s. Reconnecting in 5s...", e)
                await asyncio.sleep(5)
            except Exception as e:
                log.error("Unexpected gateway error: %s", e, exc_info=True)
                await asyncio.sleep(5)

    async def _connect_and_run(self, url: str) -> None:
        """Single connection attempt: connect, handshake, then listen."""
        connect_url = self._resume_gateway_url or url
        if self._resume_gateway_url:
            connect_url = f"{self._resume_gateway_url}?v=1&encoding=json"

        assert self._session is not None
        self._ws = await self._session.ws_connect(connect_url, max_msg_size=0)
        log.info("WebSocket connected")

        try:
            await self._event_loop()
        finally:
            self._stop_heartbeat()
            if self._ws and not self._ws.closed:
                await self._ws.close()

    async def _event_loop(self) -> None:
        """Main receive loop: read messages and handle opcodes."""
        if self._ws is None:
            raise GatewayNotConnected("WebSocket connection not established")

        async for msg in self._ws:
            if msg.type == aiohttp.WSMsgType.TEXT:
                payload = GatewayPayload.from_json(msg.data)
                await self._handle_payload_task(payload)

            elif msg.type == aiohttp.WSMsgType.BINARY:
                payload = GatewayPayload.from_json(msg.data.decode("utf-8"))
                await self._handle_payload_task(payload)

            elif msg.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSING):
                log.info("WebSocket closing")
                break

            elif msg.type == aiohttp.WSMsgType.ERROR:
                log.error("WebSocket error: %s", self._ws.exception())
                break

        if self._ws.close_code:
            await self._handle_close_code(self._ws.close_code)

    async def _handle_payload_task(self, payload: GatewayPayload) -> None:
        """Simple wrapper function to ensure that gateway payload handling is done truly asynchronously"""
        task = asyncio.create_task(self._handle_payload(payload))
        self._tasks.append(task)
        task.add_done_callback(lambda t: self._tasks.remove(t))
        task.add_done_callback(lambda t: t.result())
    
    async def _handle_payload(self, payload: GatewayPayload) -> None:
        """Route an incoming payload by opcode."""
        log.debug("Received: %s", payload)

        match payload.op:
            case GatewayOpcode.HELLO:
                self._heartbeat_interval = payload.d["heartbeat_interval"] / 1000.0
                self._start_heartbeat()

                if self._session_id:
                    await self._send_resume()
                else:
                    await self._send_identify()

            case GatewayOpcode.HEARTBEAT_ACK:
                self._last_heartbeat_ack = True
                log.debug("Heartbeat ACK received")

            case GatewayOpcode.HEARTBEAT:
                await self._send_heartbeat()

            case GatewayOpcode.DISPATCH:
                if payload.s is not None:
                    self._sequence = payload.s

                event_name = payload.t or ""
                await self._handle_dispatch(event_name, payload.d)

            case GatewayOpcode.RECONNECT:
                log.info("Gateway requested reconnect")
                if self._ws:
                    await self._ws.close()

            case GatewayOpcode.INVALID_SESSION:
                resumable = payload.d if isinstance(payload.d, bool) else False
                log.warning("Invalid session (resumable=%s)", resumable)
                if not resumable:
                    self._session_id = None
                    self._sequence = None
                    self._resume_gateway_url = None
                await asyncio.sleep(1 + (5 * (not resumable)))
                if self._ws:
                    await self._ws.close()

    async def _handle_dispatch(self, event_name: str, data: Any) -> None:
        """Handle a DISPATCH event (op 0)."""
        match event_name:
            case "READY":
                self._session_id = data["session_id"]
                self._resume_gateway_url = data.get("resume_gateway_url")
                log.info(
                    "READY: session=%s, user=%s",
                    self._session_id,
                    data["user"].get("username", "?"),
                )

            case "RESUMED":
                log.info("Successfully resumed session")

        await self._dispatch(event_name, data)

    async def _handle_close_code(self, code: int) -> None:
        """Handle a WebSocket close code from the gateway."""
        log.info("Gateway closed with code %d", code)

        try:
            close_code = GatewayCloseCode(code)
            if not close_code.is_reconnectable:
                log.error(
                    "Fatal close code %d (%s), not reconnecting", code, close_code.name
                )
                self._is_closed = True
        except ValueError:
            # Unknown close code, try to reconnect
            log.warning("Unknown close code %d, attempting reconnect", code)

    # =========================================================================
    # Sending
    # =========================================================================

    async def _send(self, payload: GatewayPayload) -> None:
        """Send a payload over the WebSocket."""
        if self._ws is None or self._ws.closed:
            log.warning("Tried to send on closed WebSocket")
            return
        raw = payload.to_json()
        log.debug("Sending: %s", payload)
        await self._ws.send_str(raw)

    async def _send_identify(self) -> None:
        """Send the IDENTIFY payload to start a new session."""
        payload = GatewayPayload(
            op=GatewayOpcode.IDENTIFY,
            d={
                "token": self._token,
                "intents": int(self._intents),
                "properties": {
                    "os": sys.platform,
                    "browser": "fluxer.py",
                    "device": "fluxer.py",
                },
            },
        )
        await self._send(payload)
        log.info("Sent IDENTIFY")

    async def _send_resume(self) -> None:
        """Send the RESUME payload to continue an existing session."""
        payload = GatewayPayload(
            op=GatewayOpcode.RESUME,
            d={
                "token": self._token,
                "session_id": self._session_id,
                "seq": self._sequence,
            },
        )
        await self._send(payload)
        log.info("Sent RESUME (session=%s, seq=%s)", self._session_id, self._sequence)

    async def _send_heartbeat(self) -> None:
        """Send a heartbeat to keep the connection alive."""
        payload = GatewayPayload(op=GatewayOpcode.HEARTBEAT, d=self._sequence)
        await self._send(payload)

    # =========================================================================
    # Heartbeat loop
    # =========================================================================

    def _start_heartbeat(self) -> None:
        self._stop_heartbeat()
        self._last_heartbeat_ack = True
        self._heartbeat_task = asyncio.create_task(self._heartbeat_loop())

    def _stop_heartbeat(self) -> None:
        if self._heartbeat_task:
            self._heartbeat_task.cancel()
            self._heartbeat_task = None

    async def _heartbeat_loop(self) -> None:
        """Periodically send heartbeats. If we miss an ACK, reconnect."""
        try:
            import random

            await asyncio.sleep(self._heartbeat_interval * random.random())
            while True:
                if not self._last_heartbeat_ack:
                    log.warning(
                        "Missed heartbeat ACK, closing connection to trigger reconnect"
                    )
                    if self._ws and not self._ws.closed:
                        await self._ws.close(code=4000)
                    return

                self._last_heartbeat_ack = False
                await self._send_heartbeat()
                await asyncio.sleep(self._heartbeat_interval)
        except asyncio.CancelledError:
            pass

    # =========================================================================
    # Lifecycle
    # =========================================================================

    async def close(self) -> None:
        """Gracefully close the gateway connection."""
        self._is_closed = True
        self._stop_heartbeat()
        if self._ws and not self._ws.closed:
            await self._ws.close(code=1000)
        if self._session and not self._session.closed:
            await self._session.close()

    async def update_presence(
        self,
        *,
        status: str = "online",
        activity_name: str | None = None,
        activity_type: int = 0,
    ) -> None:
        """Update the bot's presence/status."""
        activity = None
        if activity_name:
            activity = {"name": activity_name, "type": activity_type}

        payload = GatewayPayload(
            op=GatewayOpcode.PRESENCE_UPDATE,
            d={
                "since": None,
                "activities": [activity] if activity else [],
                "status": status,
                "afk": False,
            },
        )
        await self._send(payload)



from __future__ import annotations

import asyncio
import logging
from typing import Any, Callable, Coroutine

from .enums import Intents
from .gateway import Gateway
from .http import HTTPClient
from .models import Channel, Guild, Message, User, UserProfile, Webhook

log = logging.getLogger(__name__)

# Type alias for event handlers
EventHandler = Callable[..., Coroutine[Any, Any, None]]


class Client:
    """Low-level client that connects to Fluxer and dispatches events.

    This gives you full control over the gateway lifecycle.
    For most bots, use the Bot subclass instead.
    """

    def __init__(self, *, intents: Intents = Intents.default()) -> None:
        self.intents = intents
        self._http: HTTPClient | None = None
        self._gateway: Gateway | None = None
        self._event_handlers: dict[str, list[EventHandler]] = {}
        self._user: User | None = None
        self._guilds: dict[str, Guild] = {}
        self._channels: dict[str, Channel] = {}
        self._closed: bool = False

    @property
    def user(self) -> User | None:
        """The bot user, available after the READY event."""
        return self._user

    @property
    def guilds(self) -> list[Guild]:
        """List of guilds the bot is in (populated from READY + GUILD_CREATE)."""
        return list(self._guilds.values())

    # =========================================================================
    # Event registration
    # =========================================================================

    def event(self, func: EventHandler) -> EventHandler:
        """Decorator to register an event handler.

        The function name determines the event:
            @bot.event
            async def on_message(message):
                ...

        Supported events (mapped from gateway dispatch names):
            on_ready       -> READY
            on_message      -> MESSAGE_CREATE
            on_message_edit -> MESSAGE_UPDATE
            on_message_delete -> MESSAGE_DELETE
            on_guild_join   -> GUILD_CREATE
            on_guild_remove -> GUILD_DELETE
            on_member_join  -> GUILD_MEMBER_ADD
            on_member_remove -> GUILD_MEMBER_REMOVE
            ... and any other gateway event as on_{lowercase_name}
        """
        event_name = func.__name__
        if not event_name.startswith("on_"):
            raise ValueError(f"Event handler must start with 'on_', got '{event_name}'")

        if event_name not in self._event_handlers:
            self._event_handlers[event_name] = []
        self._event_handlers[event_name].append(func)
        return func

    def on(self, event_name: str) -> Callable[[EventHandler], EventHandler]:
        """Decorator to register an event handler with an explicit name.

        Usage:
            @bot.on("message")
            async def handle_msg(message):
                ...
        """

        def decorator(func: EventHandler) -> EventHandler:
            key = f"on_{event_name}"
            if key not in self._event_handlers:
                self._event_handlers[key] = []
            self._event_handlers[key].append(func)
            return func

        return decorator

    # =========================================================================
    # Event dispatching
    # =========================================================================

    async def _dispatch(self, event_name: str, data: Any) -> None:
        """Called by the Gateway when a dispatch event is received.

        This method:
        1. Parses the raw data into model objects
        2. Updates internal caches
        3. Fires matching user event handlers
        """
        # Map gateway event names to handler names and parse data
        match event_name:
            case "READY":
                self._user = User.from_data(data["user"], self._http)
                # Process guilds from READY
                for guild_data in data.get("guilds", []):
                    guild = Guild.from_data(guild_data, self._http)
                    self._guilds[guild.id] = guild
                await self._fire("on_ready")

            case "MESSAGE_CREATE":
                message = self._parse_message(data)
                await self._fire("on_message", message)

            case "MESSAGE_UPDATE":
                message = self._parse_message(data)
                await self._fire("on_message_edit", message)

            case "MESSAGE_DELETE":
                await self._fire("on_message_delete", data)

            case "GUILD_CREATE":
                guild = Guild.from_data(data, self._http)
                self._guilds[guild.id] = guild
                # Cache channels from guild
                for ch_data in data.get("channels", []):
                    ch = Channel.from_data(ch_data, self._http)
                    self._channels[ch.id] = ch
                await self._fire("on_guild_join", guild)

            case "GUILD_DELETE":
                guild_id = data.get("id", "")
                guild = self._guilds.pop(guild_id, None)
                await self._fire("on_guild_remove", guild or data)

            case "GUILD_MEMBER_ADD":
                await self._fire("on_member_join", data)

            case "GUILD_MEMBER_REMOVE":
                await self._fire("on_member_remove", data)

            case "CHANNEL_CREATE":
                channel = Channel.from_data(data, self._http)
                self._channels[channel.id] = channel
                await self._fire("on_channel_create", channel)

            case "CHANNEL_UPDATE":
                channel = Channel.from_data(data, self._http)
                self._channels[channel.id] = channel
                await self._fire("on_channel_update", channel)

            case "CHANNEL_DELETE":
                channel = Channel.from_data(data, self._http)
                self._channels.pop(channel.id, None)
                await self._fire("on_channel_delete", channel)

            case "RESUMED":
                await self._fire("on_resumed")

            case _:
                # Unknown/unhandled event — fire a generic handler
                handler_name = f"on_{event_name.lower()}"
                await self._fire(handler_name, data)

    def _parse_message(self, data: dict[str, Any]) -> Message:
        """Parse message data and attach cached channel reference."""
        msg = Message.from_data(data, self._http)
        # Attach cached channel
        cached_channel = self._channels.get(msg.channel_id)
        if cached_channel:
            msg._channel = cached_channel
        return msg

    async def _fire(self, event_name: str, *args: Any) -> None:
        """Fire all registered handlers for an event."""
        handlers = self._event_handlers.get(event_name, [])
        for handler in handlers:
            try:
                await handler(*args)
            except Exception:
                log.exception("Error in event handler '%s'", event_name)

    # =========================================================================
    # HTTP convenience methods
    # =========================================================================

    async def fetch_channel(self, channel_id: str) -> Channel:
        """Fetch a channel from the API (not cache)."""
        assert self._http is not None
        data = await self._http.get_channel(channel_id)
        ch = Channel.from_data(data, self._http)
        self._channels[ch.id] = ch
        return ch

    async def fetch_guild(self, guild_id: str) -> Guild:
        """Fetch a guild from the API."""
        assert self._http is not None
        data = await self._http.get_guild(guild_id)
        guild = Guild.from_data(data, self._http)
        self._guilds[guild.id] = guild
        return guild

    async def fetch_user(self, user_id: str) -> User:
        """Fetch a user from the API."""
        assert self._http is not None
        data = await self._http.get_user(user_id)
        return User.from_data(data, self._http)

    async def fetch_user_profile(
        self, user_id: str, *, guild_id: str | None = None
    ) -> UserProfile:
        """Fetch a user's full profile from the API.

        This returns additional profile information like bio, pronouns, banner, etc.
        that is not included in the basic user object.

        Args:
            user_id: The user ID to fetch
            guild_id: Optional guild ID for guild-specific profile data

        Returns:
            UserProfile containing the user and their profile information
        """
        assert self._http is not None
        data = await self._http.get_user_profile(user_id, guild_id=guild_id)
        return UserProfile.from_data(data, self._http)

    async def fetch_webhook(self, webhook_id: str) -> Webhook:
        """Fetch a webhook from the API."""
        assert self._http is not None
        data = await self._http.get_webhook(webhook_id)
        return Webhook.from_data(data, self._http)

    async def fetch_channel_webhooks(self, channel_id: str) -> list[Webhook]:
        """Fetch all webhooks for a channel."""
        assert self._http is not None
        data = await self._http.get_channel_webhooks(channel_id)
        return [Webhook.from_data(w, self._http) for w in data]

    async def fetch_guild_webhooks(self, guild_id: str) -> list[Webhook]:
        """Fetch all webhooks for a guild."""
        assert self._http is not None
        data = await self._http.get_guild_webhooks(guild_id)
        return [Webhook.from_data(w, self._http) for w in data]

    async def create_webhook(
        self, channel_id: str, *, name: str, avatar: str | None = None
    ) -> Webhook:
        """Create a webhook in a channel."""
        assert self._http is not None
        data = await self._http.create_webhook(channel_id, name=name, avatar=avatar)
        return Webhook.from_data(data, self._http)

    # =========================================================================
    # Connection lifecycle
    # =========================================================================

    async def start(self, token: str) -> None:
        """Connect to Fluxer and start receiving events (async version).

        Use this if you're managing your own event loop.
        """
        self._http = HTTPClient(token)
        self._gateway = Gateway(
            http_client=self._http,
            token=token,
            intents=self.intents,
            dispatch=self._dispatch,
        )

        try:
            await self._gateway.connect()
        finally:
            await self.close()

    async def close(self) -> None:
        """Disconnect from the gateway and clean up resources."""
        self._closed = True
        if self._gateway:
            await self._gateway.close()
        if self._http:
            await self._http.close()

    def run(self, token: str) -> None:
        """Blocking call that connects to Fluxer and runs the bot.

        This is the simplest way to start your bot:
            bot.run("your_token_here")

        It creates an event loop, calls start(), and handles cleanup.
        """

        async def _runner() -> None:
            try:
                await self.start(token)
            except KeyboardInterrupt:
                pass
            finally:
                if not self._closed:
                    await self.close()

        try:
            asyncio.run(_runner())
        except KeyboardInterrupt:
            log.info("Bot stopped by KeyboardInterrupt")


class Bot(Client):
    """Extended Client with common bot conveniences.

    Adds prefix command support and other bot-specific features.
    This is the recommended class for most bot use cases.
    """

    def __init__(
        self,
        *,
        command_prefix: str = "!",
        intents: Intents = Intents.default(),
    ) -> None:
        super().__init__(intents=intents)
        self.command_prefix = command_prefix
        self._commands: dict[str, EventHandler] = {}

        # Auto-register the command dispatcher
        @self.event
        async def on_message(message: Message) -> None:
            await self._process_commands(message)

    def command(
        self, name: str | None = None
    ) -> Callable[[EventHandler], EventHandler]:
        """Decorator to register a prefix command.

        Usage:
            @bot.command()
            async def ping(message):
                await message.reply("Pong!")

            @bot.command(name="hello")
            async def greet(message):
                await message.reply(f"Hello, {message.author}!")
        """

        def decorator(func: EventHandler) -> EventHandler:
            cmd_name = name or func.__name__
            self._commands[cmd_name] = func
            self._commands = dict(sorted(self._commands.items(), key=lambda kv: len(kv[0]), reverse=True)) # sorts the dictionary in reverse key length order
            return func

        return decorator

    async def _process_commands(self, message: Message) -> None:
        """Check if a message matches a registered command and invoke it."""
        if message.author.bot:
            return
        if not message.content.startswith(self.command_prefix):
            return

        # Parse command name and args
        content = message.content[len(self.command_prefix) :]
        for cmd, handler in self._commands.items():
            if content.startswith(cmd):
                if handler:
                    try:
                        await handler(message)
                    except Exception:
                        log.exception("Error in command '%s'", cmd)
                    break



from __future__ import annotations

from typing import Any


class FluxerException(Exception):
    """Base exception for all fluxer.py errors."""


# =============================================================================
# HTTP Errors
# =============================================================================


class HTTPException(FluxerException):
    """Raised when an HTTP request to the Fluxer API fails."""

    def __init__(
        self,
        status: int,
        code: str,
        message: str,
        errors: list[dict[str, Any]] | None = None,
    ):
        self.status = status
        self.code = code
        self.message = message
        self.errors = errors or []
        super().__init__(f"{status} {code}: {message}")


class BadRequest(HTTPException):
    """400 Bad Request — invalid input."""


class Unauthorized(HTTPException):
    """401 Unauthorized — bad or missing token."""


class Forbidden(HTTPException):
    """403 Forbidden — missing permissions."""


class NotFound(HTTPException):
    """404 Not Found — resource doesn't exist."""


class RateLimited(HTTPException):
    """429 Too Many Requests — slow down."""

    def __init__(self, retry_after: float, **kwargs: Any):
        self.retry_after = retry_after
        super().__init__(
            status=429,
            code="RATE_LIMITED",
            message=f"Rate limited, retry after {retry_after}s",
        )


# =============================================================================
# Gateway Errors
# =============================================================================


class GatewayException(FluxerException):
    """Base for gateway/WebSocket errors."""


class GatewayNotConnected(GatewayException):
    """Raised when trying to use the gateway before it's connected."""


class ReconnectRequested(GatewayException):
    """The gateway has requested we reconnect."""


class SessionInvalid(GatewayException):
    """The session is invalid and cannot be resumed."""

    def __init__(self, resumable: bool = False):
        self.resumable = resumable
        super().__init__(f"Invalid session (resumable={resumable})")


# =============================================================================
# Client Errors
# =============================================================================


class LoginFailure(FluxerException):
    """Raised when the bot token is invalid."""


# Map HTTP status codes to exception classes
_STATUS_MAP: dict[int, type[HTTPException]] = {
    400: BadRequest,
    401: Unauthorized,
    403: Forbidden,
    404: NotFound,
    429: RateLimited,
}


def http_exception_from_status(
    status: int, code: str, message: str, **kwargs: Any
) -> HTTPException:
    """Factory to create the right HTTPException subclass for a status code."""
    cls = _STATUS_MAP.get(status, HTTPException)
    if cls is RateLimited:
        return RateLimited(retry_after=kwargs.get("retry_after", 0.0))
    return cls(status=status, code=code, message=message, errors=kwargs.get("errors"))



__version__ = "0.2.0"
__title__ = "fluxer.py"
__author__ = "Emil"
__license__ = "MIT"

# Core classes
from .client import Bot, Client
from .enums import ChannelType, GatewayCloseCode, GatewayOpcode, Intents
from .http import HTTPClient

# Errors
from .errors import (
    BadRequest,
    FluxerException,
    Forbidden,
    GatewayException,
    GatewayNotConnected,
    HTTPException,
    LoginFailure,
    NotFound,
    RateLimited,
    Unauthorized,
)

# Models
from .models import Channel, Embed, Emoji, Guild, GuildMember, Message, User, UserProfile, Webhook

# Utilities
from .utils import datetime_to_snowflake, snowflake_to_datetime

__all__ = [
    # Client
    "Bot",
    "Client",
    "HTTPClient",
    # Enums
    "ChannelType",
    "GatewayCloseCode",
    "GatewayOpcode",
    "Intents",
    # Errors
    "BadRequest",
    "FluxerException",
    "Forbidden",
    "GatewayException",
    "GatewayNotConnected",
    "HTTPException",
    "LoginFailure",
    "NotFound",
    "RateLimited",
    "Unauthorized",
    # Models
    "Channel",
    "Embed",
    "Emoji",
    "Guild",
    "GuildMember",
    "Message",
    "User",
    "UserProfile",
    "Webhook",
    # Utils
    "datetime_to_snowflake",
    "snowflake_to_datetime",
]



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from fluxer.models.user import User

if TYPE_CHECKING:
    from ..http import HTTPClient
    from .message import Message


@dataclass(slots=True)
class Webhook:
    """Represents a Fluxer webhook."""

    id: int
    guild_id: int
    channel_id: int
    user: User
    name: str
    avatar: str | None
    token: str

    _http: HTTPClient | None = field(default=None, repr=False)

    @classmethod
    def from_data(
        cls,
        data: dict[str, Any],
        http: HTTPClient | None = None,
        *,
        guild_id: int | None = None,
    ) -> Webhook:
        """Construct a Webhook from raw API data.

        Args:
            data: Raw webhook object from the API.
            http: HTTPClient for making further requests.
            guild_id: Override guild_id if not present in data.

        Returns:
            A new Webhook instance.
        """
        return cls(
            id=int(data["id"]),
            guild_id=guild_id or int(data["guild_id"]),
            channel_id=int(data["channel_id"]),
            user=User.from_data(data["user"], http),
            name=data["name"],
            avatar=data.get("avatar", None),
            token=data["token"],
            _http=http,
        )

    async def edit(
        self,
        *,
        name: str | None = None,
        avatar: str | None = None,
        channel_id: int | None = None,
    ) -> Webhook:
        """Edit this webhook.

        Args:
            name: New webhook name.
            avatar: New avatar (base64 data URI).
            channel_id: Move webhook to a different channel.

        Returns:
            The updated Webhook.
        """
        if not self._http:
            raise RuntimeError("Cannot edit webhook without HTTPClient")

        data = await self._http.modify_webhook(
            self.id, name=name, avatar=avatar, channel_id=channel_id
        )
        return Webhook.from_data(data, self._http)

    async def send(
        self,
        content: str | None = None,
        *,
        embeds: list[dict[str, Any]] | None = None,
        username: str | None = None,
        avatar_url: str | None = None,
        wait: bool = False,
    ) -> Message | None:
        """Send a message with this webhook.

        Args:
            content: Text content of the message.
            embeds: List of embed dicts to include.
            username: Override the webhook's default name.
            avatar_url: Override the webhook's default avatar.
            wait: If True, returns the created Message.

        Returns:
            The created Message if wait=True, otherwise None.
        """
        if not self._http:
            raise RuntimeError("Cannot send with webhook without HTTPClient")

        data = await self._http.execute_webhook(
            self.id,
            self.token,
            content=content,
            embeds=embeds,
            username=username,
            avatar_url=avatar_url,
            wait=wait,
        )
        if data is not None:
            from .message import Message

            return Message.from_data(data, self._http)
        return None

    async def delete(self, *, reason: str | None = None) -> None:
        """Delete this webhook.

        Args:
            reason: Reason for deletion (shows in audit log)

        Raises:
            Forbidden: You don't have permission to delete this webhook
            NotFound: Webhook doesn't exist
            HTTPException: Deleting the webhook failed
        """
        if not self._http:
            raise RuntimeError("Cannot delete webhook without HTTPClient")

        await self._http.delete_webhook(self.id, reason=reason)



from .channel import Channel
from .embed import Embed
from .emoji import Emoji
from .guild import Guild
from .member import GuildMember
from .message import Message
from .profile import UserProfile
from .user import User
from .webhook import Webhook

__all__ = [
    "Channel",
    "Embed",
    "Emoji",
    "Guild",
    "GuildMember",
    "Message",
    "UserProfile",
    "User",
    "Webhook",
]
